from mads_datasets.datasets.basicdatasets import (
    FacesDataset,
    PdDataset,
    PolarsDataset,
    TextDataset,
)
from mads_datasets.datasets.torchdatasets import (
    ImgDataset,
    MNISTDataset,
    SunspotDataset,
    TensorDataset,
    TSDataset,
)

__all__ = [
    "PdDataset",
    "PolarsDataset",
    "FacesDataset",
    "TextDataset",
    "SunspotDataset",
    "TensorDataset",
    "MNISTDataset",
    "ImgDataset",
    "TSDataset",
]
